"""
Chromaticity Coordinates of the Light Sources
=============================================

Defines the chromaticity coordinates of the light sources.

The following light sources are available:

-   *RIT* *PointerData.xls* spreadsheet light sources: Natural,
    Philips TL-84, T8 Luxline Plus White, SA, SC, T8 Polylux 3000,
    T8 Polylux 4000, Thorn Kolor-rite
-   *NIST* *NIST CQS simulation 7.4.xls* spreadsheet traditional light sources:
    Cool White FL, Daylight FL, HPS, Incandescent, LPS, Mercury,
    Metal Halide, Neodimium Incandescent, Super HPS, Triphosphor FL
-   *NIST* *NIST CQS simulation 7.4.xls* spreadsheet LED light sources:
    3-LED-1 (457/540/605), 3-LED-2 (473/545/616), 3-LED-2 Yellow,
    3-LED-3 (465/546/614), 3-LED-4 (455/547/623), 4-LED No Yellow,
    4-LED Yellow, 4-LED-1 (461/526/576/624), 4-LED-2 (447/512/573/627),
    Luxeon WW 2880, PHOS-1, PHOS-2, PHOS-3, PHOS-4,
    Phosphor LED YAG
-   *NIST* *NIST CQS simulation 7.4.xls* spreadsheet Philips light sources:
    60 A/W (Soft White), C100S54 (HPS), C100S54C (HPS),
    F32T8/TL830 (Triphosphor), F32T8/TL835 (Triphosphor),
    F32T8/TL841 (Triphosphor), F32T8/TL850 (Triphosphor),
    F32T8/TL865 /PLUS (Triphosphor), F34/CW/RS/EW (Cool White FL),
    F34T12/LW/RS /EW, F34T12WW/RS /EW (Warm White FL),
    F40/C50 (Broadband FL), F40/C75 (Broadband FL),
    F40/CWX (Broadband FL), F40/DX (Broadband FL), F40/DXTP (Delux FL),
    F40/N (Natural FL), H38HT-100 (Mercury), H38JA-100/DX (Mercury DX),
    MHC100/U/MP /3K, MHC100/U/MP /4K, SDW-T 100W/LV (Super HPS)
-   Common light sources: Kinoton 75P

References
----------
-   :cite:`Houston2015a` : Borer, T. (2017). Private Discussion with Mansencal,
    T. and Shaw, N.
-   :cite:`Ohno2008a` : Ohno, Yoshiro, & Davis, W. (2008). NIST CQS simulation
    (Version 7.4) [Computer software].
    https://drive.google.com/file/d/1PsuU6QjUJjCX6tQyCud6ul2Tbs8rYWW9/view?\
usp=sharing
-   :cite:`Pointer1980a` : Pointer, M. R. (1980). Pointer's Gamut Data.
    http://www.cis.rit.edu/research/mcsl2/online/PointerData.xls
"""

from __future__ import annotations

import numpy as np

from colour.utilities import CaseInsensitiveMapping

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CCS_LIGHT_SOURCES_RIT_STANDARD_OBSERVER_2_DEGREE_CIE1931",
    "CCS_LIGHT_SOURCES_RIT_STANDARD_OBSERVER_10_DEGREE_CIE1964",
    "CCS_LIGHT_SOURCES_NIST_TRADITIONAL_STANDARD_OBSERVER_2_DEGREE_CIE1931",
    "CCS_LIGHT_SOURCES_NIST_TRADITIONAL_STANDARD_OBSERVER_10_DEGREE_CIE1964",
    "CCS_LIGHT_SOURCES_NIST_LED_STANDARD_OBSERVER_2_DEGREE_CIE1931",
    "CCS_LIGHT_SOURCES_NIST_LED_STANDARD_OBSERVER_10_DEGREE_CIE1964",
    "CCS_LIGHT_SOURCES_NIST_PHILIPS_STANDARD_OBSERVER_2_DEGREE_CIE1931",
    "CCS_LIGHT_SOURCES_NIST_PHILIPS_STANDARD_OBSERVER_10_DEGREE_CIE1964",
    "CCS_LIGHT_SOURCES_COMMON_STANDARD_OBSERVER_2_DEGREE_CIE1931",
    "CCS_LIGHT_SOURCES_COMMON_STANDARD_OBSERVER_10_DEGREE_CIE1964",
    "CCS_LIGHT_SOURCES",
]

CCS_LIGHT_SOURCES_RIT_STANDARD_OBSERVER_2_DEGREE_CIE1931: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "Natural": np.array([0.381585730647787, 0.359224138274067]),
        "Philips TL-84": np.array([0.378413599970988, 0.379290254544090]),
        "SA": np.array([0.447573030734154, 0.407438137156467]),
        "SC": np.array([0.310056734303928, 0.316145704789204]),
        "T8 Luxline Plus White": np.array(
            [0.410492204086250, 0.388932529676840]
        ),
        "T8 Polylux 3000": np.array([0.431706082207185, 0.413877736072647]),
        "T8 Polylux 4000": np.array([0.379219473139794, 0.384469085577631]),
        "Thorn Kolor-rite": np.array([0.381919124282806, 0.374309261641251]),
    }
)
"""
Chromaticity coordinates of the light sources from the *RIT* *PointerData.xls*
spreadsheet for the *CIE 1931 2 Degree Standard Observer*.

Warnings
--------
The chromaticity coordinates have been calculated from *PointerData.xls*
spreadsheet which doesn't mention the data source thus the light source names
cannot be accurately verified.

References
----------
:cite:`Pointer1980a`
"""

CCS_LIGHT_SOURCES_RIT_STANDARD_OBSERVER_10_DEGREE_CIE1964: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "Natural": np.array([0.384870991183035, 0.353869223366545]),
        "Philips TL-84": np.array([0.383592002892950, 0.373922741815762]),
        "SA": np.array([0.451176803594070, 0.405936046781591]),
        "SC": np.array([0.310388637415649, 0.319050651220986]),
        "T8 Luxline Plus White": np.array(
            [0.416946978831203, 0.380991426462756]
        ),
        "T8 Polylux 3000": np.array([0.439038926288670, 0.404554330124715]),
        "T8 Polylux 4000": np.array([0.385115161872875, 0.377800928395769]),
        "Thorn Kolor-rite": np.array([0.385533929282467, 0.370840492090948]),
    }
)
"""
Chromaticity coordinates of the light sources from the *RIT* *PointerData.xls*
spreadsheet for the *CIE 1964 10 Degree Standard Observer*.
"""

CCS_LIGHT_SOURCES_NIST_TRADITIONAL_STANDARD_OBSERVER_2_DEGREE_CIE1931: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "Cool White FL": np.array([0.369256318971281, 0.372549878176631]),
        "Daylight FL": np.array([0.312662993963651, 0.331985688793009]),
        "HPS": np.array([0.521677696062816, 0.417971177117239]),
        "Incandescent": np.array([0.450730217519680, 0.408046128945005]),
        "LPS": np.array([0.575151311365165, 0.424232234924905]),
        "Mercury": np.array([0.392018457637112, 0.383777071984453]),
        "Metal Halide": np.array([0.372544558972793, 0.385603925927588]),
        "Neodimium Incandescent": np.array(
            [0.447398697052100, 0.395008601248268]
        ),
        "Super HPS": np.array([0.470061659271846, 0.406116584248741]),
        "Triphosphor FL": np.array([0.413163268257275, 0.396422053758680]),
    }
)
"""
Chromaticity coordinates of the traditional light sources from the *NIST*
*NIST CQS simulation 7.4.xls* spreadsheet for the
*CIE 1931 2 Degree Standard Observer*.

References
----------
:cite:`Ohno2008a`
"""

CCS_LIGHT_SOURCES_NIST_TRADITIONAL_STANDARD_OBSERVER_10_DEGREE_CIE1964: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "Cool White FL": np.array([0.376715047518455, 0.364576802118673]),
        "Daylight FL": np.array([0.317395878738965, 0.330780819136676]),
        "HPS": np.array([0.531764495177513, 0.408752715284645]),
        "Incandescent": np.array([0.454365604973572, 0.406573684216774]),
        "LPS": np.array([0.589960045887891, 0.410039954112109]),
        "Mercury": np.array([0.401266412873755, 0.364732538221183]),
        "Metal Halide": np.array([0.378786167751226, 0.377496928504661]),
        "Neodimium Incandescent": np.array(
            [0.447516717156694, 0.396734151368497]
        ),
        "Super HPS": np.array([0.473859567146135, 0.401381825309197]),
        "Triphosphor FL": np.array([0.418591963931736, 0.388947713332192]),
    }
)
"""
Chromaticity coordinates of the traditional light sources from the *NIST*
*NIST CQS simulation 7.4.xls* spreadsheet for the
*CIE 1964 10 Degree Standard Observer*.
"""

CCS_LIGHT_SOURCES_NIST_LED_STANDARD_OBSERVER_2_DEGREE_CIE1931: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "3-LED-1 (457/540/605)": np.array(
            [0.417057686949170, 0.396262457986602]
        ),
        "3-LED-2 (473/545/616)": np.array(
            [0.417060475566006, 0.396268120523418]
        ),
        "3-LED-2 Yellow": np.array([0.436563079184047, 0.443649619298676]),
        "3-LED-3 (465/546/614)": np.array(
            [0.380460502184482, 0.376772001481922]
        ),
        "3-LED-4 (455/547/623)": np.array(
            [0.417067943691045, 0.396276280071757]
        ),
        "4-LED No Yellow": np.array([0.417060589301332, 0.396268153712350]),
        "4-LED Yellow": np.array([0.417069637940463, 0.396276766014859]),
        "4-LED-1 (461/526/576/624)": np.array(
            [0.417067615440556, 0.396275056779587]
        ),
        "4-LED-2 (447/512/573/627)": np.array(
            [0.417071570560054, 0.396278745130373]
        ),
        "Luxeon WW 2880": np.array([0.459088527920913, 0.432916480607903]),
        "PHOS-1": np.array([0.436443167801164, 0.404616033549917]),
        "PHOS-2": np.array([0.452704462198571, 0.437584543052711]),
        "PHOS-3": np.array([0.436899870751359, 0.404037372134463]),
        "PHOS-4": np.array([0.436936023906427, 0.404113558278629]),
        "Phosphor LED YAG": np.array([0.307761817314310, 0.325268939239941]),
    }
)
"""
Chromaticity coordinates of the LED light sources from the *NIST*
*NIST CQS simulation 7.4.xls* spreadsheet for the
*CIE 1931 2 Degree Standard Observer*.
"""

CCS_LIGHT_SOURCES_NIST_LED_STANDARD_OBSERVER_10_DEGREE_CIE1964: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "3-LED-1 (457/540/605)": np.array(
            [0.425099988926548, 0.389451349911075]
        ),
        "3-LED-2 (473/545/616)": np.array(
            [0.422222118774217, 0.401298495594226]
        ),
        "3-LED-2 Yellow": np.array([0.446222216139125, 0.441646464276087]),
        "3-LED-3 (465/546/614)": np.array(
            [0.387470465801936, 0.376404716015666]
        ),
        "3-LED-4 (455/547/623)": np.array(
            [0.422865464107041, 0.388772240171637]
        ),
        "4-LED No Yellow": np.array([0.419807532952439, 0.399465294930377]),
        "4-LED Yellow": np.array([0.422720601750053, 0.390284663473479]),
        "4-LED-1 (461/526/576/624)": np.array(
            [0.423899783323037, 0.394170886226971]
        ),
        "4-LED-2 (447/512/573/627)": np.array(
            [0.421571042053867, 0.394089741928601]
        ),
        "Luxeon WW 2880": np.array([0.466639299623263, 0.430817417218051]),
        "PHOS-1": np.array([0.440120001281140, 0.403135783393416]),
        "PHOS-2": np.array([0.461487398870558, 0.436150294667024]),
        "PHOS-3": np.array([0.440892655302172, 0.408662264402299]),
        "PHOS-4": np.array([0.441760443951475, 0.407267478268879]),
        "Phosphor LED YAG": np.array([0.312807834772696, 0.334180937864035]),
    }
)
"""
Chromaticity coordinates of the LED light sources from the *NIST*
*NIST CQS simulation 7.4.xls* spreadsheet for the
*CIE 1964 10 Degree Standard Observer*.
"""

CCS_LIGHT_SOURCES_NIST_PHILIPS_STANDARD_OBSERVER_2_DEGREE_CIE1931: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "60 A/W (Soft White)": np.array(
            [0.450730217519680, 0.408046128945005]
        ),
        "C100S54 (HPS)": np.array([0.529231515407657, 0.411370164988427]),
        "C100S54C (HPS)": np.array([0.502380414374839, 0.415877299905475]),
        "F32T8/TL830 (Triphosphor)": np.array(
            [0.443250764475753, 0.409523700296928]
        ),
        "F32T8/TL835 (Triphosphor)": np.array(
            [0.407150274569933, 0.393172743482571]
        ),
        "F32T8/TL841 (Triphosphor)": np.array(
            [0.385376686681605, 0.390370762102806]
        ),
        "F32T8/TL850 (Triphosphor)": np.array(
            [0.343768910392287, 0.358447436104108]
        ),
        "F32T8/TL865 /PLUS (Triphosphor)": np.array(
            [0.316368879615201, 0.345320790143017]
        ),
        "F34/CW/RS/EW (Cool White FL)": np.array(
            [0.377250931364378, 0.393087658636060]
        ),
        "F34T12/LW/RS /EW": np.array([0.378863642993776, 0.394960629979820]),
        "F34T12WW/RS /EW (Warm White FL)": np.array(
            [0.438466967656789, 0.408635441565706]
        ),
        "F40/C50 (Broadband FL)": np.array(
            [0.345836574973021, 0.361724450389430]
        ),
        "F40/C75 (Broadband FL)": np.array(
            [0.299966663385220, 0.316582165804824]
        ),
        "F40/CWX (Broadband FL)": np.array(
            [0.375037045754214, 0.360543952129462]
        ),
        "F40/DX (Broadband FL)": np.array(
            [0.311922310746537, 0.342802103417329]
        ),
        "F40/DXTP (Delux FL)": np.array(
            [0.313066543826958, 0.342225714484412]
        ),
        "F40/N (Natural FL)": np.array([0.376878697365115, 0.354153458302878]),
        "H38HT-100 (Mercury)": np.array(
            [0.311200590193641, 0.382944245857018]
        ),
        "H38JA-100/DX (Mercury DX)": np.array(
            [0.389791630360359, 0.373394688931767]
        ),
        "MHC100/U/MP /3K": np.array([0.428581768670222, 0.388168915678330]),
        "MHC100/U/MP /4K": np.array([0.373145253482762, 0.371366990216717]),
        "SDW-T 100W/LV (Super HPS)": np.array(
            [0.472339157938672, 0.407106330880316]
        ),
    }
)
"""
Chromaticity coordinates of the Philips light sources from the *NIST*
*NIST CQS simulation 7.4.xls* spreadsheet for the
*CIE 1931 2 Degree Standard Observer*.
"""

CCS_LIGHT_SOURCES_NIST_PHILIPS_STANDARD_OBSERVER_10_DEGREE_CIE1964: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "60 A/W (Soft White)": np.array(
            [0.454365604973572, 0.406573684216774]
        ),
        "C100S54 (HPS)": np.array([0.538554605063010, 0.402575827972962]),
        "C100S54C (HPS)": np.array([0.509663059970892, 0.409064508209193]),
        "F32T8/TL830 (Triphosphor)": np.array(
            [0.448795219301811, 0.403574636091678]
        ),
        "F32T8/TL835 (Triphosphor)": np.array(
            [0.412082534290652, 0.388001071127592]
        ),
        "F32T8/TL841 (Triphosphor)": np.array(
            [0.390908619219527, 0.385290559992705]
        ),
        "F32T8/TL850 (Triphosphor)": np.array(
            [0.347882431257452, 0.355845742210551]
        ),
        "F32T8/TL865 /PLUS (Triphosphor)": np.array(
            [0.320698199593768, 0.343871441043854]
        ),
        "F34/CW/RS/EW (Cool White FL)": np.array(
            [0.386514853545337, 0.382843326097814]
        ),
        "F34T12/LW/RS /EW": np.array([0.389628909159399, 0.382074721889904]),
        "F34T12WW/RS /EW (Warm White FL)": np.array(
            [0.448395377616960, 0.395666643335296]
        ),
        "F40/C50 (Broadband FL)": np.array(
            [0.349880827196884, 0.360661316491439]
        ),
        "F40/C75 (Broadband FL)": np.array(
            [0.301988533872761, 0.318479025875818]
        ),
        "F40/CWX (Broadband FL)": np.array(
            [0.378502309910296, 0.356371890168937]
        ),
        "F40/DX (Broadband FL)": np.array(
            [0.316783037559153, 0.341749269085077]
        ),
        "F40/DXTP (Delux FL)": np.array(
            [0.318774745065791, 0.339798825605488]
        ),
        "F40/N (Natural FL)": np.array([0.378833157741751, 0.350724402658646]),
        "H38HT-100 (Mercury)": np.array(
            [0.326260627082484, 0.360001095895205]
        ),
        "H38JA-100/DX (Mercury DX)": np.array(
            [0.397058597517533, 0.356532431806974]
        ),
        "MHC100/U/MP /3K": np.array([0.431422986591898, 0.380642213887539]),
        "MHC100/U/MP /4K": np.array([0.375707105948115, 0.366156465779779]),
        "SDW-T 100W/LV (Super HPS)": np.array(
            [0.476461908192661, 0.402288012403575]
        ),
    }
)
"""
Chromaticity coordinates of the Philips light sources from the *NIST*
*NIST CQS simulation 7.4.xls* spreadsheet for the
*CIE 1964 10 Degree Standard Observer*.
"""

CCS_LIGHT_SOURCES_COMMON_STANDARD_OBSERVER_2_DEGREE_CIE1931: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {"Kinoton 75P": np.array([0.315252413629716, 0.332870794805328])}
)
"""
Chromaticity coordinates of the common light sources for the
*CIE 1931 2 Degree Standard Observer*.

References
----------
:cite:`Houston2015a`
"""

CCS_LIGHT_SOURCES_COMMON_STANDARD_OBSERVER_10_DEGREE_CIE1964: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {"Kinoton 75P": np.array([0.317086642148234, 0.336222428041514])}
)
"""
Chromaticity coordinates of the common light sources for the
*CIE 1964 10 Degree Standard Observer*.
"""

CCS_LIGHT_SOURCES: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "CIE 1931 2 Degree Standard Observer": CaseInsensitiveMapping(
            CCS_LIGHT_SOURCES_RIT_STANDARD_OBSERVER_2_DEGREE_CIE1931
        ),
        "CIE 1964 10 Degree Standard Observer": CaseInsensitiveMapping(
            CCS_LIGHT_SOURCES_RIT_STANDARD_OBSERVER_10_DEGREE_CIE1964
        ),
    }
)
CCS_LIGHT_SOURCES.__doc__ = """
Chromaticity coordinates of the light sources.

Aliases:

-   'cie_2_1931': 'CIE 1931 2 Degree Standard Observer'
-   'cie_10_1964': 'CIE 1964 10 Degree Standard Observer'
"""
CCS_LIGHT_SOURCES["cie_2_1931"] = CCS_LIGHT_SOURCES[
    "CIE 1931 2 Degree Standard Observer"
]
CCS_LIGHT_SOURCES["cie_10_1964"] = CCS_LIGHT_SOURCES[
    "CIE 1964 10 Degree Standard Observer"
]

CCS_LIGHT_SOURCES["CIE 1931 2 Degree Standard Observer"].update(
    CCS_LIGHT_SOURCES_NIST_TRADITIONAL_STANDARD_OBSERVER_2_DEGREE_CIE1931
)
CCS_LIGHT_SOURCES["CIE 1964 10 Degree Standard Observer"].update(
    CCS_LIGHT_SOURCES_NIST_TRADITIONAL_STANDARD_OBSERVER_10_DEGREE_CIE1964
)

CCS_LIGHT_SOURCES["CIE 1931 2 Degree Standard Observer"].update(
    CCS_LIGHT_SOURCES_NIST_LED_STANDARD_OBSERVER_2_DEGREE_CIE1931
)
CCS_LIGHT_SOURCES["CIE 1964 10 Degree Standard Observer"].update(
    CCS_LIGHT_SOURCES_NIST_LED_STANDARD_OBSERVER_10_DEGREE_CIE1964
)

CCS_LIGHT_SOURCES["CIE 1931 2 Degree Standard Observer"].update(
    CCS_LIGHT_SOURCES_NIST_PHILIPS_STANDARD_OBSERVER_2_DEGREE_CIE1931
)
CCS_LIGHT_SOURCES["CIE 1964 10 Degree Standard Observer"].update(
    CCS_LIGHT_SOURCES_NIST_PHILIPS_STANDARD_OBSERVER_10_DEGREE_CIE1964
)

CCS_LIGHT_SOURCES["CIE 1931 2 Degree Standard Observer"].update(
    CCS_LIGHT_SOURCES_COMMON_STANDARD_OBSERVER_2_DEGREE_CIE1931
)
CCS_LIGHT_SOURCES["CIE 1964 10 Degree Standard Observer"].update(
    CCS_LIGHT_SOURCES_COMMON_STANDARD_OBSERVER_10_DEGREE_CIE1964
)



"""
Michaelis-Menten Kinetics
=========================

Implements support for *Michaelis-Menten* kinetics, a model of enzyme kinetics:

-   :func:`colour.biochemistry.reaction_rate_MichaelisMenten_Michaelis1913`
-   :func:`colour.biochemistry.reaction_rate_MichaelisMenten_Abebe2017`
-   :func:`colour.biochemistry.REACTION_RATE_MICHAELISMENTEN_METHODS`
-   :func:`colour.biochemistry.reaction_rate_MichaelisMenten`
-   :func:`colour.biochemistry.\
substrate_concentration_MichaelisMenten_Michaelis1913`
-   :func:`colour.biochemistry.\
substrate_concentration_MichaelisMenten_Abebe2017`
-   :func:`colour.biochemistry.SUBSTRATE_CONCENTRATION_MICHAELISMENTEN_METHODS`
-   :func:`colour.biochemistry.substrate_concentration_MichaelisMenten`

References
----------
-   :cite:`Abebe2017a` : Abebe, M. A., Pouli, T., Larabi, M.-C., & Reinhard,
    E. (2017). Perceptual Lightness Modeling for High-Dynamic-Range Imaging.
    ACM Transactions on Applied Perception, 15(1), 1-19. doi:10.1145/3086577
-   :cite:`Wikipedia2003d` : Wikipedia. (2003). Michaelis-Menten kinetics.
    Retrieved April 29, 2017, from
    https://en.wikipedia.org/wiki/Michaelis%E2%80%93Menten_kinetics
"""

from __future__ import annotations

from colour.hints import (
    Any,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Literal,
    Union,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    as_float,
    as_float_array,
    filter_kwargs,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "reaction_rate_MichaelisMenten_Michaelis1913",
    "reaction_rate_MichaelisMenten_Abebe2017",
    "REACTION_RATE_MICHAELISMENTEN_METHODS",
    "reaction_rate_MichaelisMenten",
    "substrate_concentration_MichaelisMenten_Michaelis1913",
    "substrate_concentration_MichaelisMenten_Abebe2017",
    "SUBSTRATE_CONCENTRATION_MICHAELISMENTEN_METHODS",
    "substrate_concentration_MichaelisMenten",
]


def reaction_rate_MichaelisMenten_Michaelis1913(
    S: FloatingOrArrayLike,
    V_max: FloatingOrArrayLike,
    K_m: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Describe the rate of enzymatic reactions, by relating reaction rate
    :math:`v` to concentration of a substrate :math:`S`.

    Parameters
    ----------
    S
        Concentration of a substrate :math:`S`.
    V_max
        Maximum rate :math:`V_{max}` achieved by the system, at saturating
        substrate concentration.
    K_m
        Substrate concentration :math:`K_m` at which the reaction rate is
        half of :math:`V_{max}`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Reaction rate :math:`v`.

    References
    ----------
    :cite:`Wikipedia2003d`

    Examples
    --------
    >>> reaction_rate_MichaelisMenten(0.5, 2.5, 0.8)  # doctest: +ELLIPSIS
    0.9615384...
    """

    S = as_float_array(S)
    V_max = as_float_array(V_max)
    K_m = as_float_array(K_m)

    v = (V_max * S) / (K_m + S)

    return as_float(v)


def reaction_rate_MichaelisMenten_Abebe2017(
    S: FloatingOrArrayLike,
    V_max: FloatingOrArrayLike,
    K_m: FloatingOrArrayLike,
    b_m: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Describe the rate of enzymatic reactions, by relating reaction rate
    :math:`v` to concentration of a substrate :math:`S` according to the
    modified *Michaelis-Menten* kinetics equation as given by
    *Abebe, Pouli, Larabi and Reinhard (2017)*.

    Parameters
    ----------
    S
        Concentration of a substrate :math:`S` (or
        :math:`(\\cfrac{Y}{Y_n})^{\\epsilon}`).
    V_max
        Maximum rate :math:`V_{max}` (or :math:`a_m`) achieved by the system,
        at saturating substrate concentration.
    K_m
        Substrate concentration :math:`K_m` (or :math:`c_m`) at which the
        reaction rate is half of :math:`V_{max}`.
    b_m
        Bias factor :math:`b_m`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Reaction rate :math:`v`.

    References
    ----------
    :cite:`Abebe2017a`

    Examples
    --------
    >>> reaction_rate_MichaelisMenten_Abebe2017(0.5, 1.448, 0.635, 0.813)
    ... # doctest: +ELLIPSIS
    0.6951512...
    """

    S = as_float_array(S)
    V_max = as_float_array(V_max)
    K_m = as_float_array(K_m)
    b_m = as_float_array(b_m)

    v = (V_max * S) / (b_m * S + K_m)

    return as_float(v)


REACTION_RATE_MICHAELISMENTEN_METHODS: CaseInsensitiveMapping = (
    CaseInsensitiveMapping(
        {
            "Michaelis 1913": reaction_rate_MichaelisMenten_Michaelis1913,
            "Abebe 2017": reaction_rate_MichaelisMenten_Abebe2017,
        }
    )
)
REACTION_RATE_MICHAELISMENTEN_METHODS.__doc__ = """
Supported *Michaelis-Menten* kinetics reaction rate equation computation
methods.

References
----------
:cite:`Wikipedia2003d`, :cite:`Abebe2017a`
"""


def reaction_rate_MichaelisMenten(
    S: FloatingOrArrayLike,
    V_max: FloatingOrArrayLike,
    K_m: FloatingOrArrayLike,
    method: Union[
        Literal["Michaelis 1913", "Abebe 2017"], str
    ] = "Michaelis 1913",
    **kwargs: Any,
) -> FloatingOrNDArray:
    """
    Describe the rate of enzymatic reactions, by relating reaction rate
    :math:`v` to concentration of a substrate :math:`S` according to given
    method.

    Parameters
    ----------
    S
        Concentration of a substrate :math:`S`.
    V_max
        Maximum rate :math:`V_{max}` achieved by the system, at saturating
        substrate concentration.
    K_m
        Substrate concentration :math:`K_m` at which the reaction rate is
        half of :math:`V_{max}`.
    method
        Computation method.

    Other Parameters
    ----------------
    b_m
        {:func:`colour.biochemistry.reaction_rate_MichaelisMenten_Abebe2017`},
        Bias factor :math:`b_m`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Reaction rate :math:`v`.

    References
    ----------
    :cite:`Wikipedia2003d`, :cite:`Abebe2017a`

    Examples
    --------
    >>> reaction_rate_MichaelisMenten(0.5, 2.5, 0.8)  # doctest: +ELLIPSIS
    0.9615384...
    >>> reaction_rate_MichaelisMenten(
    ... 0.5, 2.5, 0.8, method='Abebe 2017', b_m=0.813)  # doctest: +ELLIPSIS
    1.0360547...
    """

    method = validate_method(method, REACTION_RATE_MICHAELISMENTEN_METHODS)

    function = REACTION_RATE_MICHAELISMENTEN_METHODS[method]

    return function(S, V_max, K_m, **filter_kwargs(function, **kwargs))


def substrate_concentration_MichaelisMenten_Michaelis1913(
    v: FloatingOrArrayLike,
    V_max: FloatingOrArrayLike,
    K_m: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Describe the rate of enzymatic reactions, by relating concentration of a
    substrate :math:`S` to reaction rate :math:`v`.

    Parameters
    ----------
    v
        Reaction rate :math:`v`.
    V_max
        Maximum rate :math:`V_{max}` achieved by the system, at saturating
        substrate concentration.
    K_m
        Substrate concentration :math:`K_m` at which the reaction rate is
        half of :math:`V_{max}`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Concentration of a substrate :math:`S`.

    References
    ----------
    :cite:`Wikipedia2003d`

    Examples
    --------
    >>> substrate_concentration_MichaelisMenten(0.961538461538461, 2.5, 0.8)
    ... # doctest: +ELLIPSIS
    0.4999999...
    """

    v = as_float_array(v)
    V_max = as_float_array(V_max)
    K_m = as_float_array(K_m)

    S = (v * K_m) / (V_max - v)

    return as_float(S)


def substrate_concentration_MichaelisMenten_Abebe2017(
    v: FloatingOrArrayLike,
    V_max: FloatingOrArrayLike,
    K_m: FloatingOrArrayLike,
    b_m: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Describe the rate of enzymatic reactions, by relating concentration of a
    substrate :math:`S` to reaction rate :math:`v` according to the modified
    *Michaelis-Menten* kinetics equation as given by
    *Abebe, Pouli, Larabi and Reinhard (2017)*.

    Parameters
    ----------
    v
        Reaction rate :math:`v`.
    V_max
        Maximum rate :math:`V_{max}` (or :math:`a_m`) achieved by the system,
        at saturating substrate concentration.
    K_m
        Substrate concentration :math:`K_m` (or :math:`c_m`) at which the
        reaction rate is half of :math:`V_{max}`.
    b_m
        Bias factor :math:`b_m`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Concentration of a substrate :math:`S`.

    References
    ----------
    :cite:`Abebe2017a`

    Examples
    --------
    >>> substrate_concentration_MichaelisMenten_Abebe2017(
    ...     0.695151224195871, 1.448, 0.635, 0.813)  # doctest: +ELLIPSIS
    0.4999999...
    """

    v = as_float_array(v)
    V_max = as_float_array(V_max)
    K_m = as_float_array(K_m)
    b_m = as_float_array(b_m)

    S = (v * K_m) / (V_max - b_m * v)

    return as_float(S)


SUBSTRATE_CONCENTRATION_MICHAELISMENTEN_METHODS: (
    CaseInsensitiveMapping
) = CaseInsensitiveMapping(
    {
        "Michaelis 1913": substrate_concentration_MichaelisMenten_Michaelis1913,
        "Abebe 2017": substrate_concentration_MichaelisMenten_Abebe2017,
    }
)
SUBSTRATE_CONCENTRATION_MICHAELISMENTEN_METHODS.__doc__ = """
Supported *Michaelis-Menten* kinetics substrate concentration equation
computation methods.

References
----------
:cite:`Wikipedia2003d`, :cite:`Abebe2017a`
"""


def substrate_concentration_MichaelisMenten(
    v: FloatingOrArrayLike,
    V_max: FloatingOrArrayLike,
    K_m: FloatingOrArrayLike,
    method: Union[
        Literal["Michaelis 1913", "Abebe 2017"], str
    ] = "Michaelis 1913",
    **kwargs: Any,
) -> FloatingOrNDArray:
    """
    Describe the rate of enzymatic reactions, by relating concentration of a
    substrate :math:`S` to reaction rate :math:`v` according to given method.

    Parameters
    ----------
    v
        Reaction rate :math:`v`.
    V_max
        Maximum rate :math:`V_{max}` achieved by the system, at saturating
        substrate concentration.
    K_m
        Substrate concentration :math:`K_m` at which the reaction rate is
        half of :math:`V_{max}`.
    method
        Computation method.

    Other Parameters
    ----------------
    b_m
        {:func:`colour.biochemistry.\
substrate_concentration_MichaelisMenten_Abebe2017`},
        Bias factor :math:`b_m`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Concentration of a substrate :math:`S`.

    References
    ----------
    :cite:`Wikipedia2003d`, :cite:`Abebe2017a`

    Examples
    --------
    >>> substrate_concentration_MichaelisMenten(0.961538461538461, 2.5, 0.8)
    ... # doctest: +ELLIPSIS
    0.4999999...
    >>> substrate_concentration_MichaelisMenten(
    ... 1.036054703688355, 2.5, 0.8, method='Abebe 2017', b_m=0.813)
    ... # doctest: +ELLIPSIS
    0.5000000...
    """

    method = validate_method(
        method, SUBSTRATE_CONCENTRATION_MICHAELISMENTEN_METHODS
    )

    function = SUBSTRATE_CONCENTRATION_MICHAELISMENTEN_METHODS[method]

    return function(v, V_max, K_m, **filter_kwargs(function, **kwargs))



"""Defines the unit tests for the :mod:`colour.difference.cam16_ucs` module."""

import unittest

from colour.difference.tests.test_cam02_ucs import TestDelta_E_Luo2006

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestDelta_E_Li2017",
]


class TestDelta_E_Li2017(TestDelta_E_Luo2006):
    """
    Define :func:`colour.difference.cam16_ucs.delta_E_Li2017` definition unit
    tests methods.

    Notes
    -----
    -   :func:`colour.difference.cam16_ucs.delta_E_Li2017` is a wrapper
        of :func:`colour.difference.cam02_ucs.delta_E_Luo2006` and thus
        currently adopts the same unittests.
    """


if __name__ == "__main__":
    unittest.main()



"""
:math:`\\Delta E^*_{ab}` - Delta E Colour Difference
====================================================

Defines the :math:`\\Delta E^*_{ab}` colour difference computation objects:

The following attributes and methods are available:

-   :attr:`colour.difference.JND_CIE1976`
-   :func:`colour.difference.delta_E_CIE1976`
-   :func:`colour.difference.delta_E_CIE1994`
-   :func:`colour.difference.delta_E_CIE2000`
-   :func:`colour.difference.delta_E_CMC`

References
----------
-   :cite:`Lindbloom2003c` : Lindbloom, B. (2003). Delta E (CIE 1976).
    Retrieved February 24, 2014, from
    http://brucelindbloom.com/Eqn_DeltaE_CIE76.html
-   :cite:`Lindbloom2009e` : Lindbloom, B. (2009). Delta E (CIE 2000).
    Retrieved February 24, 2014, from
    http://brucelindbloom.com/Eqn_DeltaE_CIE2000.html
-   :cite:`Lindbloom2009f` : Lindbloom, B. (2009). Delta E (CMC). Retrieved
    February 24, 2014, from http://brucelindbloom.com/Eqn_DeltaE_CMC.html
-   :cite:`Lindbloom2011a` : Lindbloom, B. (2011). Delta E (CIE 1994).
    Retrieved February 24, 2014, from
    http://brucelindbloom.com/Eqn_DeltaE_CIE94.html
-   :cite:`Melgosa2013b` : Melgosa, M. (2013). CIE / ISO new standard:
    CIEDE2000. http://www.color.org/events/colorimetry/\
Melgosa_CIEDE2000_Workshop-July4.pdf
-   :cite:`Mokrzycki2011` : Mokrzycki, W., & Tatol, M. (2011). Color difference
    Delta E - A survey. Machine Graphics and Vision, 20, 383-411.
"""

from __future__ import annotations

import numpy as np

from colour.algebra import euclidean_distance
from colour.hints import ArrayLike, Boolean, Floating, FloatingOrNDArray
from colour.utilities import as_float, to_domain_100, tsplit
from colour.utilities.documentation import (
    DocstringFloat,
    is_documentation_building,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "JND_CIE1976",
    "delta_E_CIE1976",
    "delta_E_CIE1994",
    "delta_E_CIE2000",
    "delta_E_CMC",
]

JND_CIE1976 = 2.3
if is_documentation_building():  # pragma: no cover
    JND_CIE1976 = DocstringFloat(JND_CIE1976)
    JND_CIE1976.__doc__ = """
Just Noticeable Difference (JND) according to *CIE 1976* colour difference
formula, i.e. Euclidean distance in *CIE L\\*a\\*b\\** colourspace.

Notes
-----
A standard observer sees the difference in colour as follows:

-   0 < :math:`\\Delta E^*_{ab}` < 1 : Observer does not notice the difference.
-   1 < :math:`\\Delta E^*_{ab}` < 2 : Only experienced observer can notice the
    difference.
-   2 < :math:`\\Delta E^*_{ab}` < 3:5 : Unexperienced observer also notices
    the difference.
-   3:5 < :math:`\\Delta E^*_{ab}` < 5 : Clear difference in colour is noticed.
-   5 < :math:`\\Delta E^*_{ab}` : Observer notices two different colours.

References
----------
:cite:`Mokrzycki2011`
"""


def delta_E_CIE1976(Lab_1: ArrayLike, Lab_2: ArrayLike) -> FloatingOrNDArray:
    """
    Return the difference :math:`\\Delta E_{76}` between two given
    *CIE L\\*a\\*b\\** colourspace arrays using *CIE 1976* recommendation.

    Parameters
    ----------
    Lab_1
        *CIE L\\*a\\*b\\** colourspace array 1.
    Lab_2
        *CIE L\\*a\\*b\\** colourspace array 2.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Colour difference :math:`\\Delta E_{76}`.

    Notes
    -----
    +------------+-----------------------+-------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**     |
    +============+=======================+===================+
    | ``Lab_1``  | ``L_1`` : [0, 100]    | ``L_1`` : [0, 1]  |
    |            |                       |                   |
    |            | ``a_1`` : [-100, 100] | ``a_1`` : [-1, 1] |
    |            |                       |                   |
    |            | ``b_1`` : [-100, 100] | ``b_1`` : [-1, 1] |
    +------------+-----------------------+-------------------+
    | ``Lab_2``  | ``L_2`` : [0, 100]    | ``L_2`` : [0, 1]  |
    |            |                       |                   |
    |            | ``a_2`` : [-100, 100] | ``a_2`` : [-1, 1] |
    |            |                       |                   |
    |            | ``b_2`` : [-100, 100] | ``b_2`` : [-1, 1] |
    +------------+-----------------------+-------------------+

    References
    ----------
    :cite:`Lindbloom2003c`

    Examples
    --------
    >>> Lab_1 = np.array([100.00000000, 21.57210357, 272.22819350])
    >>> Lab_2 = np.array([100.00000000, 426.67945353, 72.39590835])
    >>> delta_E_CIE1976(Lab_1, Lab_2)  # doctest: +ELLIPSIS
    451.7133019...
    """

    d_E = euclidean_distance(to_domain_100(Lab_1), to_domain_100(Lab_2))

    return d_E


def delta_E_CIE1994(
    Lab_1: ArrayLike, Lab_2: ArrayLike, textiles: Boolean = False
) -> FloatingOrNDArray:
    """
    Return the difference :math:`\\Delta E_{94}` between two given
    *CIE L\\*a\\*b\\** colourspace arrays using *CIE 1994* recommendation.

    Parameters
    ----------
    Lab_1
        *CIE L\\*a\\*b\\** colourspace array 1.
    Lab_2
        *CIE L\\*a\\*b\\** colourspace array 2.
    textiles
        Textiles application specific parametric factors,
        :math:`k_L=2,\\ k_C=k_H=1,\\ k_1=0.048,\\ k_2=0.014` weights are used
        instead of :math:`k_L=k_C=k_H=1,\\ k_1=0.045,\\ k_2=0.015`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Colour difference :math:`\\Delta E_{94}`.

    Notes
    -----
    +------------+-----------------------+-------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**     |
    +============+=======================+===================+
    | ``Lab_1``  | ``L_1`` : [0, 100]    | ``L_1`` : [0, 1]  |
    |            |                       |                   |
    |            | ``a_1`` : [-100, 100] | ``a_1`` : [-1, 1] |
    |            |                       |                   |
    |            | ``b_1`` : [-100, 100] | ``b_1`` : [-1, 1] |
    +------------+-----------------------+-------------------+
    | ``Lab_2``  | ``L_2`` : [0, 100]    | ``L_2`` : [0, 1]  |
    |            |                       |                   |
    |            | ``a_2`` : [-100, 100] | ``a_2`` : [-1, 1] |
    |            |                       |                   |
    |            | ``b_2`` : [-100, 100] | ``b_2`` : [-1, 1] |
    +------------+-----------------------+-------------------+

    -   *CIE 1994* colour differences are not symmetrical: difference between
        ``Lab_1`` and ``Lab_2`` may not be the same as difference between
        ``Lab_2`` and ``Lab_1`` thus one colour must be understood to be the
        reference against which a sample colour is compared.

    References
    ----------
    :cite:`Lindbloom2011a`

    Examples
    --------
    >>> Lab_1 = np.array([100.00000000, 21.57210357, 272.22819350])
    >>> Lab_2 = np.array([100.00000000, 426.67945353, 72.39590835])
    >>> delta_E_CIE1994(Lab_1, Lab_2)  # doctest: +ELLIPSIS
    83.7792255...
    >>> delta_E_CIE1994(Lab_1, Lab_2, textiles=True)  # doctest: +ELLIPSIS
    88.3355530...
    """

    L_1, a_1, b_1 = tsplit(to_domain_100(Lab_1))
    L_2, a_2, b_2 = tsplit(to_domain_100(Lab_2))

    k_1 = 0.048 if textiles else 0.045
    k_2 = 0.014 if textiles else 0.015
    k_L = 2 if textiles else 1
    k_C = 1
    k_H = 1

    C_1 = np.hypot(a_1, b_1)
    C_2 = np.hypot(a_2, b_2)

    s_L = 1
    s_C = 1 + k_1 * C_1
    s_H = 1 + k_2 * C_1

    delta_L = L_1 - L_2
    delta_C = C_1 - C_2
    delta_A = a_1 - a_2
    delta_B = b_1 - b_2

    delta_H = np.sqrt(delta_A**2 + delta_B**2 - delta_C**2)

    L = (delta_L / (k_L * s_L)) ** 2
    C = (delta_C / (k_C * s_C)) ** 2
    H = (delta_H / (k_H * s_H)) ** 2

    d_E = np.sqrt(L + C + H)

    return as_float(d_E)


def delta_E_CIE2000(
    Lab_1: ArrayLike, Lab_2: ArrayLike, textiles: Boolean = False
) -> FloatingOrNDArray:
    """
    Return the difference :math:`\\Delta E_{00}` between two given
    *CIE L\\*a\\*b\\** colourspace arrays using *CIE 2000* recommendation.

    Parameters
    ----------
    Lab_1
        *CIE L\\*a\\*b\\** colourspace array 1.
    Lab_2
        *CIE L\\*a\\*b\\** colourspace array 2.
    textiles
        Textiles application specific parametric factors.
        :math:`k_L=2,\\ k_C=k_H=1` weights are used instead of
        :math:`k_L=k_C=k_H=1`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Colour difference :math:`\\Delta E_{00}`.

    Notes
    -----
    +------------+-----------------------+-------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**     |
    +============+=======================+===================+
    | ``Lab_1``  | ``L_1`` : [0, 100]    | ``L_1`` : [0, 1]  |
    |            |                       |                   |
    |            | ``a_1`` : [-100, 100] | ``a_1`` : [-1, 1] |
    |            |                       |                   |
    |            | ``b_1`` : [-100, 100] | ``b_1`` : [-1, 1] |
    +------------+-----------------------+-------------------+
    | ``Lab_2``  | ``L_2`` : [0, 100]    | ``L_2`` : [0, 1]  |
    |            |                       |                   |
    |            | ``a_2`` : [-100, 100] | ``a_2`` : [-1, 1] |
    |            |                       |                   |
    |            | ``b_2`` : [-100, 100] | ``b_2`` : [-1, 1] |
    +------------+-----------------------+-------------------+

    -   Parametric factors :math:`k_L=k_C=k_H=1` weights under
        *reference conditions*:

        -   Illumination: D65 source
        -   Illuminance: 1000 lx
        -   Observer: Normal colour vision
        -   Background field: Uniform, neutral gray with :math:`L^*=50`
        -   Viewing mode: Object
        -   Sample size: Greater than 4 degrees
        -   Sample separation: Direct edge contact
        -   Sample colour-difference magnitude: Lower than 5.0
            :math:`\\Delta E_{00}`
        -   Sample structure: Homogeneous (without texture)

    References
    ----------
    :cite:`Lindbloom2009e`, :cite:`Melgosa2013b`

    Examples
    --------
    >>> Lab_1 = np.array([100.00000000, 21.57210357, 272.22819350])
    >>> Lab_2 = np.array([100.00000000, 426.67945353, 72.39590835])
    >>> delta_E_CIE2000(Lab_1, Lab_2)  # doctest: +ELLIPSIS
    94.0356490...
    >>> Lab_2 = np.array([50.00000000, 426.67945353, 72.39590835])
    >>> delta_E_CIE2000(Lab_1, Lab_2)  # doctest: +ELLIPSIS
    100.8779470...
    >>> delta_E_CIE2000(Lab_1, Lab_2, textiles=True)  # doctest: +ELLIPSIS
    95.7920535...
    """

    L_1, a_1, b_1 = tsplit(to_domain_100(Lab_1))
    L_2, a_2, b_2 = tsplit(to_domain_100(Lab_2))

    k_L = 2 if textiles else 1
    k_C = 1
    k_H = 1

    l_bar_prime = 0.5 * (L_1 + L_2)

    c_1 = np.hypot(a_1, b_1)
    c_2 = np.hypot(a_2, b_2)

    c_bar = 0.5 * (c_1 + c_2)
    c_bar7 = c_bar**7

    g = 0.5 * (1 - np.sqrt(c_bar7 / (c_bar7 + 25**7)))

    a_1_prime = a_1 * (1 + g)
    a_2_prime = a_2 * (1 + g)
    c_1_prime = np.hypot(a_1_prime, b_1)
    c_2_prime = np.hypot(a_2_prime, b_2)
    c_bar_prime = 0.5 * (c_1_prime + c_2_prime)

    h_1_prime = np.degrees(np.arctan2(b_1, a_1_prime)) % 360
    h_2_prime = np.degrees(np.arctan2(b_2, a_2_prime)) % 360

    h_bar_prime = np.where(
        np.fabs(h_1_prime - h_2_prime) <= 180,
        0.5 * (h_1_prime + h_2_prime),
        (0.5 * (h_1_prime + h_2_prime + 360)),
    )

    t = (
        1
        - 0.17 * np.cos(np.deg2rad(h_bar_prime - 30))
        + 0.24 * np.cos(np.deg2rad(2 * h_bar_prime))
        + 0.32 * np.cos(np.deg2rad(3 * h_bar_prime + 6))
        - 0.20 * np.cos(np.deg2rad(4 * h_bar_prime - 63))
    )

    h = h_2_prime - h_1_prime
    delta_h_prime = np.where(h_2_prime <= h_1_prime, h - 360, h + 360)
    delta_h_prime = np.where(np.fabs(h) <= 180, h, delta_h_prime)

    delta_L_prime = L_2 - L_1
    delta_C_prime = c_2_prime - c_1_prime
    delta_H_prime = (
        2
        * np.sqrt(c_1_prime * c_2_prime)
        * np.sin(np.deg2rad(0.5 * delta_h_prime))
    )

    s_L = 1 + (
        (0.015 * (l_bar_prime - 50) * (l_bar_prime - 50))
        / np.sqrt(20 + (l_bar_prime - 50) * (l_bar_prime - 50))
    )
    s_C = 1 + 0.045 * c_bar_prime
    s_H = 1 + 0.015 * c_bar_prime * t

    delta_theta = 30 * np.exp(
        -((h_bar_prime - 275) / 25) * ((h_bar_prime - 275) / 25)
    )

    c_bar_prime7 = c_bar_prime**7

    r_C = np.sqrt(c_bar_prime7 / (c_bar_prime7 + 25**7))
    r_T = -2 * r_C * np.sin(np.deg2rad(2 * delta_theta))

    d_E = np.sqrt(
        (delta_L_prime / (k_L * s_L)) ** 2
        + (delta_C_prime / (k_C * s_C)) ** 2
        + (delta_H_prime / (k_H * s_H)) ** 2
        + (delta_C_prime / (k_C * s_C)) * (delta_H_prime / (k_H * s_H)) * r_T
    )

    return as_float(d_E)


def delta_E_CMC(
    Lab_1: ArrayLike,
    Lab_2: ArrayLike,
    l: Floating = 2,  # noqa
    c: Floating = 1,
) -> FloatingOrNDArray:
    """
    Return the difference :math:`\\Delta E_{CMC}` between two given
    *CIE L\\*a\\*b\\** colourspace arrays using *Colour Measurement Committee*
    recommendation.

    The quasimetric has two parameters: *Lightness* (l) and *chroma* (c),
    allowing the users to weight the difference based on the ratio of l:c.
    Commonly used values are 2:1 for acceptability and 1:1 for the threshold of
    imperceptibility.

    Parameters
    ----------
    Lab_1
        *CIE L\\*a\\*b\\** colourspace array 1.
    Lab_2
        *CIE L\\*a\\*b\\** colourspace array 2.
    l
        Lightness weighting factor.
    c
        Chroma weighting factor.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Colour difference :math:`\\Delta E_{CMC}`.

    Notes
    -----
    +------------+-----------------------+-------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**     |
    +============+=======================+===================+
    | ``Lab_1``  | ``L_1`` : [0, 100]    | ``L_1`` : [0, 1]  |
    |            |                       |                   |
    |            | ``a_1`` : [-100, 100] | ``a_1`` : [-1, 1] |
    |            |                       |                   |
    |            | ``b_1`` : [-100, 100] | ``b_1`` : [-1, 1] |
    +------------+-----------------------+-------------------+
    | ``Lab_2``  | ``L_2`` : [0, 100]    | ``L_2`` : [0, 1]  |
    |            |                       |                   |
    |            | ``a_2`` : [-100, 100] | ``a_2`` : [-1, 1] |
    |            |                       |                   |
    |            | ``b_2`` : [-100, 100] | ``b_2`` : [-1, 1] |
    +------------+-----------------------+-------------------+

    References
    ----------
    :cite:`Lindbloom2009f`

    Examples
    --------
    >>> Lab_1 = np.array([100.00000000, 21.57210357, 272.22819350])
    >>> Lab_2 = np.array([100.00000000, 426.67945353, 72.39590835])
    >>> delta_E_CMC(Lab_1, Lab_2)  # doctest: +ELLIPSIS
    172.7047712...
    """

    L_1, a_1, b_1 = tsplit(to_domain_100(Lab_1))
    L_2, a_2, b_2 = tsplit(to_domain_100(Lab_2))

    c_1 = np.hypot(a_1, b_1)
    c_2 = np.hypot(a_2, b_2)
    s_l = np.where(L_1 < 16, 0.511, (0.040975 * L_1) / (1 + 0.01765 * L_1))
    s_c = 0.0638 * c_1 / (1 + 0.0131 * c_1) + 0.638
    h_1 = np.degrees(np.arctan2(b_1, a_1)) % 360

    t = np.where(
        np.logical_and(h_1 >= 164, h_1 <= 345),
        0.56 + np.fabs(0.2 * np.cos(np.deg2rad(h_1 + 168))),
        0.36 + np.fabs(0.4 * np.cos(np.deg2rad(h_1 + 35))),
    )

    c_4 = c_1 * c_1 * c_1 * c_1
    f = np.sqrt(c_4 / (c_4 + 1900))
    s_h = s_c * (f * t + 1 - f)

    delta_L = L_1 - L_2
    delta_C = c_1 - c_2
    delta_A = a_1 - a_2
    delta_B = b_1 - b_2
    delta_H2 = delta_A**2 + delta_B**2 - delta_C**2

    v_1 = delta_L / (l * s_l)
    v_2 = delta_C / (c * s_c)
    v_3 = s_h

    d_E = np.sqrt(v_1**2 + v_2**2 + (delta_H2 / (v_3 * v_3)))

    return as_float(d_E)



"""Defines the unit tests for the :mod:`colour.colorimetry.photometry` module."""

import unittest

from colour.colorimetry import (
    SDS_ILLUMINANTS,
    SDS_LIGHT_SOURCES,
    luminous_flux,
    luminous_efficiency,
    luminous_efficacy,
    sd_zeros,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestLuminousFlux",
    "TestLuminousEfficiency",
    "TestLuminousEfficacy",
]


class TestLuminousFlux(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.photometry.luminous_flux` definition unit
    tests methods.
    """

    def test_luminous_flux(self):
        """Test :func:`colour.colorimetry.photometry.luminous_flux` definition."""

        self.assertAlmostEqual(
            luminous_flux(SDS_ILLUMINANTS["FL2"].copy().normalise()),
            28588.73612977,
            places=7,
        )

        self.assertAlmostEqual(
            luminous_flux(SDS_LIGHT_SOURCES["Neodimium Incandescent"]),
            23807.65552737,
            places=7,
        )

        self.assertAlmostEqual(
            luminous_flux(SDS_LIGHT_SOURCES["F32T8/TL841 (Triphosphor)"]),
            13090.06759053,
            places=7,
        )


class TestLuminousEfficiency(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.photometry.luminous_efficiency`
    definition unit tests methods.
    """

    def test_luminous_efficiency(self):
        """
        Test :func:`colour.colorimetry.photometry.luminous_efficiency`
        definition.
        """

        self.assertAlmostEqual(
            luminous_efficiency(SDS_ILLUMINANTS["FL2"].copy().normalise()),
            0.49317624,
            places=7,
        )

        self.assertAlmostEqual(
            luminous_efficiency(SDS_LIGHT_SOURCES["Neodimium Incandescent"]),
            0.19943936,
            places=7,
        )

        self.assertAlmostEqual(
            luminous_efficiency(
                SDS_LIGHT_SOURCES["F32T8/TL841 (Triphosphor)"]
            ),
            0.51080919,
            places=7,
        )


class TestLuminousEfficacy(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.photometry.luminous_efficacy`
    definition unit tests methods.
    """

    def test_luminous_efficacy(self):
        """
        Test :func:`colour.colorimetry.photometry.luminous_efficacy`
        definition.
        """

        self.assertAlmostEqual(
            luminous_efficacy(SDS_ILLUMINANTS["FL2"].copy().normalise()),
            336.83937176,
            places=7,
        )

        self.assertAlmostEqual(
            luminous_efficacy(SDS_LIGHT_SOURCES["Neodimium Incandescent"]),
            136.21708032,
            places=7,
        )

        self.assertAlmostEqual(
            luminous_efficacy(SDS_LIGHT_SOURCES["F32T8/TL841 (Triphosphor)"]),
            348.88267549,
            places=7,
        )

        sd = sd_zeros()
        sd[555] = 1
        self.assertAlmostEqual(luminous_efficacy(sd), 683.00000000, places=7)


if __name__ == "__main__":
    unittest.main()



"""
Tristimulus Values, CIE xyY Colourspace and Chromaticity Coordinates
====================================================================

Defines the *CIE xyY* colourspace transformations:

-   :func:`colour.XYZ_to_xyY`
-   :func:`colour.xyY_to_XYZ`
-   :func:`colour.xyY_to_xy`
-   :func:`colour.xy_to_xyY`
-   :func:`colour.XYZ_to_xy`
-   :func:`colour.xy_to_XYZ`

References
----------
-   :cite:`Lindbloom2003e` : Lindbloom, B. (2003). XYZ to xyY. Retrieved
    February 24, 2014, from http://www.brucelindbloom.com/Eqn_XYZ_to_xyY.html
-   :cite:`Lindbloom2009d` : Lindbloom, B. (2009). xyY to XYZ. Retrieved
    February 24, 2014, from http://www.brucelindbloom.com/Eqn_xyY_to_XYZ.html
-   :cite:`Wikipedia2005` : Wikipedia. (2005). CIE 1931 color space. Retrieved
    February 24, 2014, from http://en.wikipedia.org/wiki/CIE_1931_color_space
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import ArrayLike, Floating, NDArray
from colour.utilities import (
    as_float_array,
    as_float_scalar,
    from_range_1,
    full,
    to_domain_1,
    tsplit,
    tstack,
    zeros,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "XYZ_to_xyY",
    "xyY_to_XYZ",
    "xy_to_xyY",
    "xyY_to_xy",
    "XYZ_to_xy",
    "xy_to_XYZ",
]


def XYZ_to_xyY(
    XYZ: ArrayLike,
    illuminant: ArrayLike = CCS_ILLUMINANTS[
        "CIE 1931 2 Degree Standard Observer"
    ]["D65"],
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to *CIE xyY* colourspace and
    reference *illuminant*.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    illuminant
        Reference *illuminant* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xyY* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``xyY``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Lindbloom2003e`, :cite:`Wikipedia2005`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_xyY(XYZ)  # doctest: +ELLIPSIS
    array([ 0.5436955...,  0.3210794...,  0.1219722...])
    """

    XYZ = to_domain_1(XYZ)
    X, Y, Z = tsplit(XYZ)
    xy_w = as_float_array(illuminant)

    XYZ_n = zeros(XYZ.shape)
    XYZ_n[..., 0:2] = xy_w

    xyY = np.where(
        np.all(XYZ == 0, axis=-1)[..., np.newaxis],
        XYZ_n,
        tstack(
            [
                X / (X + Y + Z),
                Y / (X + Y + Z),
                from_range_1(Y),
            ]
        ),
    )

    return xyY


def xyY_to_XYZ(xyY: ArrayLike) -> NDArray:
    """
    Convert from *CIE xyY* colourspace to *CIE XYZ* tristimulus values.

    Parameters
    ----------
    xyY
        *CIE xyY* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``xyY``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Lindbloom2009d`, :cite:`Wikipedia2005`

    Examples
    --------
    >>> xyY = np.array([0.54369557, 0.32107944, 0.12197225])
    >>> xyY_to_XYZ(xyY)  # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    """

    x, y, Y = tsplit(xyY)
    Y = to_domain_1(Y)

    XYZ = np.where(
        (y == 0)[..., np.newaxis],
        tstack([y, y, y]),
        tstack([x * Y / y, Y, (1 - x - y) * Y / y]),
    )

    return from_range_1(XYZ)


def xyY_to_xy(xyY: ArrayLike) -> NDArray:
    """
    Convert from *CIE xyY* colourspace to *CIE xy* chromaticity coordinates.

    ``xyY`` argument with last dimension being equal to 2 will be assumed to be
    a *CIE xy* chromaticity coordinates argument and will be returned directly
    by the definition.

    Parameters
    ----------
    xyY
        *CIE xyY* colourspace array or *CIE xy* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xy* chromaticity coordinates.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``xyY``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2005`

    Examples
    --------
    >>> xyY = np.array([0.54369557, 0.32107944, 0.12197225])
    >>> xyY_to_xy(xyY)  # doctest: +ELLIPSIS
    array([ 0.54369557...,  0.32107944...])
    >>> xy = np.array([0.54369557, 0.32107944])
    >>> xyY_to_xy(xy)  # doctest: +ELLIPSIS
    array([ 0.54369557...,  0.32107944...])
    """

    xyY = as_float_array(xyY)

    # Assuming ``xyY`` is actually a *CIE xy* chromaticity coordinates argument
    # and returning it directly.
    if xyY.shape[-1] == 2:
        return xyY

    xy = xyY[..., 0:2]

    return xy


def xy_to_xyY(xy: ArrayLike, Y: Floating = 1) -> NDArray:
    """
    Convert from *CIE xy* chromaticity coordinates to *CIE xyY* colourspace by
    extending the array last dimension with given :math:`Y` *luminance*.

    ``xy`` argument with last dimension being equal to 3 will be assumed to be
    a *CIE xyY* colourspace array argument and will be returned directly by the
    definition.

    Parameters
    ----------
    xy
        *CIE xy* chromaticity coordinates or *CIE xyY* colourspace array.
    Y
        Optional :math:`Y` *luminance* value used to construct the *CIE xyY*
        colourspace array, the default :math:`Y` *luminance* value is 1.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xyY* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``xy``     | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``xyY``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    -   This definition is a convenient object provided to implement support of
        illuminant argument *luminance* value in various :mod:`colour.models`
        package objects such as :func:`colour.Lab_to_XYZ` or
        :func:`colour.Luv_to_XYZ`.

    References
    ----------
    :cite:`Wikipedia2005`

    Examples
    --------
    >>> xy = np.array([0.54369557, 0.32107944])
    >>> xy_to_xyY(xy)  # doctest: +ELLIPSIS
    array([ 0.5436955...,  0.3210794...,  1.        ])
    >>> xy = np.array([0.54369557, 0.32107944, 1.00000000])
    >>> xy_to_xyY(xy)  # doctest: +ELLIPSIS
    array([ 0.5436955...,  0.3210794...,  1.        ])
    >>> xy = np.array([0.54369557, 0.32107944])
    >>> xy_to_xyY(xy, 100)  # doctest: +ELLIPSIS
    array([   0.5436955...,    0.3210794...,  100.        ])
    """

    xy = as_float_array(xy)
    Y = as_float_scalar(to_domain_1(Y))

    # Assuming ``xy`` is actually a *CIE xyY* colourspace array argument and
    # returning it directly.
    if xy.shape[-1] == 3:
        return xy

    x, y = tsplit(xy)

    xyY = tstack([x, y, full(x.shape, Y)])

    return from_range_1(xyY, np.array([1, 1, 100]))


def XYZ_to_xy(
    XYZ: ArrayLike,
    illuminant: ArrayLike = CCS_ILLUMINANTS[
        "CIE 1931 2 Degree Standard Observer"
    ]["D65"],
) -> NDArray:
    """
    Return the *CIE xy* chromaticity coordinates from given *CIE XYZ*
    tristimulus values.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    illuminant
        Reference *illuminant* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xy* chromaticity coordinates.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2005`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_xy(XYZ)  # doctest: +ELLIPSIS
    array([ 0.5436955...,  0.3210794...])
    """

    return xyY_to_xy(XYZ_to_xyY(XYZ, illuminant))


def xy_to_XYZ(xy: ArrayLike) -> NDArray:
    """
    Return the *CIE XYZ* tristimulus values from given *CIE xy* chromaticity
    coordinates.

    Parameters
    ----------
    xy
        *CIE xy* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``xy``     | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Wikipedia2005`

    Examples
    --------
    >>> xy = np.array([0.54369557, 0.32107944])
    >>> xy_to_XYZ(xy)  # doctest: +ELLIPSIS
    array([ 1.6933366...,  1.        ,  0.4211574...])
    """

    return xyY_to_XYZ(xy_to_xyY(xy))



"""Showcases cylindrical and spherical colour models computations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box("Cylindrical & Spherical Colour Models")

RGB = np.array([0.45620519, 0.03081071, 0.04091952])
message_box(
    f'Converting to the "HSV" colourspace from given "RGB" colourspace '
    f"values:\n\n\t{RGB}"
)
print(colour.RGB_to_HSV(RGB))

print("\n")

HSV = np.array([0.99603944, 0.93246304, 0.45620519])
message_box(
    f'Converting to the "RGB" colourspace from given "HSV" colourspace '
    f"values:\n\n\t{HSV}"
)
print(colour.HSV_to_RGB(HSV))

print("\n")

message_box(
    f'Converting to the "HSL" colourspace from given "RGB" colourspace '
    f"values:\n\n\t{RGB}"
)
print(colour.RGB_to_HSL(RGB))

print("\n")

HSL = np.array([0.99603944, 0.87347144, 0.24350795])
message_box(
    f'Converting to the "RGB" colourspace from given "HSL" colourspace '
    f"values:\n\n\t{HSL}"
)
print(colour.HSL_to_RGB(HSL))

print("\n")

message_box(
    f'Converting to the "HCL" colourspace from given "RGB" colourspace '
    f"values:\n\n\t{RGB}"
)
print(colour.RGB_to_HCL(RGB))

print("\n")

HCL = np.array([0.99603944, 0.87347144, 0.24350795])
message_box(
    f'Converting to the "RGB" colourspace from given "HCL" colourspace '
    f"values:\n\n\t{HCL}"
)
print(colour.HCL_to_RGB(HCL))

print("\n")



"""
Sharp RGB Colourspace
=====================

Defines the *Sharp RGB* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_SHARP_RGB`

References
----------
-   :cite:`Susstrunk2000` : Susstrunk, S. E., Holm, J. M., & Finlayson, G. D.
    (2000). Chromatic adaptation performance of different RGB sensors. In R.
    Eschbach & G. G. Marcu (Eds.), Photonics West 2001 - Electronic Imaging
    (Vol. 4300, Issue January, pp. 172-183). doi:10.1117/12.410788
-   :cite:`Ward2002` : Ward, G., & Eydelberg-Vileshin, E. (2002). Picture
    Perfect RGB Rendering Using Spectral Prefiltering and Sharp Color
    Primaries. Eurographics Workshop on Rendering, 117-124.
    doi:10.2312/EGWR/EGWR02/117-124
-   :cite:`Ward2016` : Borer, T. (2017). Private Discussion with Mansencal, T.
    and Shaw, N.
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    linear_function,
    normalised_primary_matrix,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_SHARP_RGB",
    "WHITEPOINT_NAME_SHARP_RGB",
    "CCS_WHITEPOINT_SHARP_RGB",
    "MATRIX_SHARP_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_SHARP_RGB",
    "RGB_COLOURSPACE_SHARP_RGB",
]

PRIMARIES_SHARP_RGB: NDArray = np.array(
    [
        [0.6898, 0.3206],
        [0.0736, 0.9003],
        [0.1166, 0.0374],
    ]
)
"""
*Sharp RGB* colourspace primaries.

Notes
-----
The primaries were originally derived from the :math:`M_{Sharp}` matrix as
given in *Ward and Eydelberg-Vileshin (2002)*:

    M_Sharp = np.array(
        [[1.2694, -0.0988, -0.1706],
         [-0.8364, 1.8006, 0.0357],
         [0.0297, -0.0315, 1.0018]])

    P, W = (
        array([[ 0.68976058,  0.32060751],
               [ 0.07358274,  0.90029055],
               [ 0.1166078 ,  0.0373923 ]]),
        array([ 0.33332778,  0.33334544]))

Private discussion with Ward (2016) confirmed he used the following primaries
and whitepoint:

    [0.6898, 0.3206, 0.0736, 0.9003, 0.1166, 0.0374, 1 / 3, 1 / 3]
"""

WHITEPOINT_NAME_SHARP_RGB: str = "E"
"""*Sharp RGB* colourspace whitepoint name."""

CCS_WHITEPOINT_SHARP_RGB: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_SHARP_RGB]
"""*Sharp RGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_SHARP_RGB_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_SHARP_RGB, CCS_WHITEPOINT_SHARP_RGB
)
"""*Sharp RGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_SHARP_RGB: NDArray = np.linalg.inv(MATRIX_SHARP_RGB_TO_XYZ)
"""*CIE XYZ* tristimulus values to *Sharp RGB* colourspace matrix."""

RGB_COLOURSPACE_SHARP_RGB: RGB_Colourspace = RGB_Colourspace(
    "Sharp RGB",
    PRIMARIES_SHARP_RGB,
    CCS_WHITEPOINT_SHARP_RGB,
    WHITEPOINT_NAME_SHARP_RGB,
    MATRIX_SHARP_RGB_TO_XYZ,
    MATRIX_XYZ_TO_SHARP_RGB,
    linear_function,
    linear_function,
)
RGB_COLOURSPACE_SHARP_RGB.__doc__ = """
*Sharp RGB* colourspace.

References
----------
:cite:`Susstrunk2000`, :cite:`Ward2002`, :cite:`Ward2016`
"""



"""
IHLS Colour Encoding
====================

Defines the :math:`IHLS` (Improved HLS) colourspace related transformations:

-   :func:`colour.RGB_to_IHLS`
-   :func:`colour.IHLS_to_RGB`

References
----------
-   :cite:`Hanbury2003` : Hanbury, A. (2003). A 3D-Polar Coordinate Colour
    Representation Well Adapted to Image Analysis. In J. Bigun & T. Gustavsson
    (Eds.), Image Analysis (pp. 804-811). Springer Berlin Heidelberg.
    ISBN:978-3-540-45103-7
"""

from __future__ import annotations

import numpy as np

from colour.algebra import vector_dot
from colour.hints import ArrayLike, NDArray
from colour.utilities import (
    from_range_1,
    to_domain_1,
    tstack,
    tsplit,
    zeros,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RGB_to_IHLS",
    "IHLS_to_RGB",
]

MATRIX_RGB_TO_YC_1_C_2: NDArray = np.array(
    [
        [0.2126, 0.7152, 0.0722],
        [1, -0.5, -0.5],
        [0, -np.sqrt(3) / 2, np.sqrt(3) / 2],
    ]
)
"""*RGB* colourspace to *YC_1C_2* colourspace matrix."""

MATRIX_YC_1_C_2_TO_RGB: NDArray = np.linalg.inv(MATRIX_RGB_TO_YC_1_C_2)
"""*YC_1C_2* colourspace to *RGB* colourspace matrix."""


def RGB_to_IHLS(RGB: ArrayLike) -> NDArray:
    """
    Convert from *RGB* colourspace to *IHLS* (Improved HLS) colourspace.

    Parameters
    ----------
    RGB
       *RGB* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *HYS* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``HYS``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Hanbury2003`

    Examples
    --------
    >>> RGB = np.array([0.45595571, 0.03039702, 0.04087245])
    >>> RGB_to_IHLS(RGB)  # doctest: +ELLIPSIS
    array([ 6.2616051...,  0.1216271...,  0.4255586...])
    """

    RGB = to_domain_1(RGB)
    R, G, B = tsplit(RGB)

    Y, C_1, C_2 = tsplit(vector_dot(MATRIX_RGB_TO_YC_1_C_2, RGB))

    C = np.sqrt(C_1**2 + C_2**2)

    acos_C_1_C_2 = zeros(C.shape)
    acos_C_1_C_2[C != 0] = np.arccos(C_1[C != 0] / C[C != 0])
    H = np.where(C_2 <= 0, acos_C_1_C_2, (np.pi * 2) - acos_C_1_C_2)

    S = np.maximum(np.maximum(R, G), B) - np.minimum(np.minimum(R, G), B)

    HYS = tstack([H, Y, S])

    return from_range_1(HYS)


def IHLS_to_RGB(HYS: ArrayLike) -> NDArray:
    """
    Convert from *IHLS* (Improved HLS) colourspace to *RGB* colourspace.

    Parameters
    ----------
    HYS
        *IHLS* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``HYS``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Hanbury2003`

    Examples
    --------
    >>> HYS = np.array([6.26160518, 0.12162712, 0.42555869])
    >>> IHLS_to_RGB(HYS)  # doctest: +ELLIPSIS
    array([ 0.4559557...,  0.0303970...,  0.0408724...])
    """

    H, Y, S = tsplit(to_domain_1(HYS))

    pi_3 = np.pi / 3

    k = np.floor(H / (pi_3))
    H_s = H - k * (pi_3)
    C = (np.sqrt(3) * S) / (2 * np.sin((2 * pi_3) - H_s))

    C_1 = C * np.cos(H)
    C_2 = -C * np.sin(H)

    RGB = vector_dot(MATRIX_YC_1_C_2_TO_RGB, tstack([Y, C_1, C_2]))

    return from_range_1(RGB)



"""
Adobe Wide Gamut RGB Colourspace
================================

Defines the *Adobe Wide Gamut RGB* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_ADOBE_WIDE_GAMUT_RGB`.

References
----------
-   :cite:`Wikipedia2004c` : Wikipedia. (2004). Wide-gamut RGB color space.
    Retrieved April 13, 2014, from
    http://en.wikipedia.org/wiki/Wide-gamut_RGB_color_space
"""

from __future__ import annotations

import numpy as np
from functools import partial

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    gamma_function,
    normalised_primary_matrix,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_ADOBE_WIDE_GAMUT_RGB",
    "WHITEPOINT_NAME_ADOBE_WIDE_GAMUT_RGB",
    "CCS_WHITEPOINT_ADOBE_WIDE_GAMUT_RGB",
    "MATRIX_ADOBE_WIDE_GAMUT_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_ADOBE_WIDE_GAMUT_RGB",
    "RGB_COLOURSPACE_ADOBE_WIDE_GAMUT_RGB",
]

PRIMARIES_ADOBE_WIDE_GAMUT_RGB: NDArray = np.array(
    [
        [0.7347, 0.2653],
        [0.1152, 0.8264],
        [0.1566, 0.0177],
    ]
)
"""*Adobe Wide Gamut RGB* colourspace primaries."""

WHITEPOINT_NAME_ADOBE_WIDE_GAMUT_RGB: str = "D50"
"""*Adobe Wide Gamut RGB* colourspace whitepoint name."""

CCS_WHITEPOINT_ADOBE_WIDE_GAMUT_RGB: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_ADOBE_WIDE_GAMUT_RGB]
"""*Adobe Wide Gamut RGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_ADOBE_WIDE_GAMUT_RGB_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_ADOBE_WIDE_GAMUT_RGB, CCS_WHITEPOINT_ADOBE_WIDE_GAMUT_RGB
)
"""*Adobe Wide Gamut RGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_ADOBE_WIDE_GAMUT_RGB: NDArray = np.linalg.inv(
    MATRIX_ADOBE_WIDE_GAMUT_RGB_TO_XYZ
)
"""*CIE XYZ* tristimulus values to *Adobe Wide Gamut RGB* colourspace matrix."""

RGB_COLOURSPACE_ADOBE_WIDE_GAMUT_RGB: RGB_Colourspace = RGB_Colourspace(
    "Adobe Wide Gamut RGB",
    PRIMARIES_ADOBE_WIDE_GAMUT_RGB,
    CCS_WHITEPOINT_ADOBE_WIDE_GAMUT_RGB,
    WHITEPOINT_NAME_ADOBE_WIDE_GAMUT_RGB,
    MATRIX_ADOBE_WIDE_GAMUT_RGB_TO_XYZ,
    MATRIX_XYZ_TO_ADOBE_WIDE_GAMUT_RGB,
    partial(gamma_function, exponent=1 / (563 / 256)),
    partial(gamma_function, exponent=563 / 256),
)
RGB_COLOURSPACE_ADOBE_WIDE_GAMUT_RGB.__doc__ = """
*Adobe Wide Gamut RGB* colourspace.

References
----------
:cite:`Wikipedia2004c`
"""



"""Defines the unit tests for the :mod:`colour.volume.pointer_gamut` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.volume import is_within_pointer_gamut
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestIsWithinPointerGamut",
]


class TestIsWithinPointerGamut(unittest.TestCase):
    """
    Define :func:`colour.volume.pointer_gamut.is_within_pointer_gamut`
    definition unit tests methods.
    """

    def test_is_within_pointer_gamut(self):
        """
        Test :func:`colour.volume.pointer_gamut.is_within_pointer_gamut`
        definition.
        """

        self.assertTrue(
            is_within_pointer_gamut(np.array([0.3205, 0.4131, 0.5100]))
        )

        self.assertFalse(
            is_within_pointer_gamut(np.array([0.0005, 0.0031, 0.0010]))
        )

        self.assertTrue(
            is_within_pointer_gamut(np.array([0.4325, 0.3788, 0.1034]))
        )

        self.assertFalse(
            is_within_pointer_gamut(np.array([0.0025, 0.0088, 0.0340]))
        )

    def test_n_dimensional_is_within_pointer_gamut(self):
        """
        Test :func:`colour.volume.pointer_gamut.is_within_pointer_gamut`
        definition n-dimensional arrays support.
        """

        a = np.array([0.3205, 0.4131, 0.5100])
        b = is_within_pointer_gamut(a)

        a = np.tile(a, (6, 1))
        b = np.tile(b, 6)
        np.testing.assert_almost_equal(is_within_pointer_gamut(a), b)

        a = np.reshape(a, (2, 3, 3))
        b = np.reshape(b, (2, 3))
        np.testing.assert_almost_equal(is_within_pointer_gamut(a), b)

    @ignore_numpy_errors
    def test_nan_is_within_pointer_gamut(self):
        """
        Test :func:`colour.volume.pointer_gamut.is_within_pointer_gamut`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            is_within_pointer_gamut(case)


if __name__ == "__main__":
    unittest.main()



"""Showcases contrast sensitivity computations."""

from pprint import pprint

import numpy as np
from scipy.optimize import fmin

import colour
from colour.hints import FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import as_float, message_box
from colour.plotting import colour_style, plot_single_function

message_box("Contrast Sensitivity Computations")

colour_style()

message_box(
    'Computing the contrast sensitivity for a spatial frequency "u" of 4, an '
    'angular size "X_0" of 60 and a retinal illuminance "E" of 65 using '
    '"Barten (1999)" method.'
)
pprint(colour.contrast_sensitivity_function(u=4, X_0=60, E=65))
pprint(
    colour.contrast.contrast_sensitivity_function_Barten1999(u=4, X_0=60, E=65)
)

print("\n")

message_box(
    "Computing the minimum detectable contrast with the assumed conditions "
    'for UHDTV applications as given in the "ITU-R BT.2246-4" "Figure 31" and '
    'using "Barten (1999)" method.'
)

settings_BT2246 = {
    "k": 3.0,
    "T": 0.1,
    "X_max": 12,
    "N_max": 15,
    "n": 0.03,
    "p": 1.2274 * 10**6,
    "phi_0": 3 * 10**-8,
    "u_0": 7,
}


def maximise_spatial_frequency(L: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Maximise the spatial frequency :math:`u` for given luminance value.

    Parameters
    ----------
    L
        Luminance value at which to maximize the spatial frequency :math:`u`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Maximised spatial frequency :math:`u`.
    """

    maximised_spatial_frequency = []
    for L_v in L:
        X_0 = 60
        d = colour.contrast.pupil_diameter_Barten1999(L_v, X_0)
        sigma = colour.contrast.sigma_Barten1999(0.5 / 60, 0.08 / 60, d)
        E = colour.contrast.retinal_illuminance_Barten1999(L_v, d, True)
        maximised_spatial_frequency.append(
            fmin(
                lambda x: (
                    -colour.contrast.contrast_sensitivity_function_Barten1999(
                        u=x, sigma=sigma, X_0=X_0, E=E, **settings_BT2246
                    )
                ),
                0,
                disp=False,
            )[0]
        )

    return as_float(np.array(maximised_spatial_frequency))


L = np.logspace(np.log10(0.01), np.log10(100), 100)
X_0 = Y_0 = 60
d = colour.contrast.barten1999.pupil_diameter_Barten1999(L, X_0, Y_0)
sigma = colour.contrast.barten1999.sigma_Barten1999(0.5 / 60, 0.08 / 60, d)
E = colour.contrast.barten1999.retinal_illuminance_Barten1999(L, d)
u = maximise_spatial_frequency(L)

pprint(
    1
    / colour.contrast_sensitivity_function(
        u=u, sigma=sigma, E=E, X_0=X_0, Y_0=Y_0, **settings_BT2246
    )
    * 2
    * (1 / 1.27)
)
pprint(
    1
    / colour.contrast.contrast_sensitivity_function_Barten1999(
        u=u, sigma=sigma, E=E, X_0=X_0, Y_0=Y_0, **settings_BT2246
    )
    * 2
    * (1 / 1.27)
)

plot_single_function(
    lambda x: (
        1
        / colour.contrast.contrast_sensitivity_function_Barten1999(
            u=u, sigma=sigma, E=E, X_0=X_0, Y_0=Y_0, **settings_BT2246
        )
        * 2
        * (1 / 1.27)
    ),
    samples=L,
    log_x=10,
    **{
        "title": "Examples of HVS Minimum Detectable Contrast Characteristics",
        "x_label": "Luminance ($cd/m^2$)",
        "y_label": "Minimum Detectable Contrast",
        "axes.grid.which": "both",
    }
)
